@@ -30,6 +30,7 @@
 #include "filesort.h"                           // filesort_free_buffers
 #include "sql_view.h"
 #include "sql_cte.h"
+#include "item_windowfunc.h"
 
 bool mysql_union(THD *thd, LEX *lex, select_result *result,
                  SELECT_LEX_UNIT *unit, ulong setup_tables_done_option)
@@ -250,7 +251,27 @@ bool select_unit::send_eof()
 
 int select_union_recursive::send_data(List<Item> &values)
 {
-  int rc= select_unit::send_data(values);
+
+  int rc;
+  bool save_abort_on_warning= thd->abort_on_warning;
+  enum_check_fields save_count_cuted_fields= thd->count_cuted_fields;
+  long save_counter;
+
+  /*
+    For recursive CTE's give warnings for wrong field info
+    However, we don't do that for CREATE TABLE ... SELECT or INSERT ... SELECT
+    as the upper level code for these handles setting of abort_on_warning
+    depending on if 'IGNORE' is used.
+  */
+  if (thd->lex->sql_command != SQLCOM_CREATE_TABLE &&
+      thd->lex->sql_command != SQLCOM_INSERT_SELECT)
+    thd->abort_on_warning= thd->is_strict_mode();
+  thd->count_cuted_fields= CHECK_FIELD_WARN;
+  save_counter= thd->get_stmt_da()->set_current_row_for_warning(++row_counter);
+  rc= select_unit::send_data(values);
+  thd->get_stmt_da()->set_current_row_for_warning(save_counter);
+  thd->count_cuted_fields= save_count_cuted_fields;
+  thd->abort_on_warning= save_abort_on_warning;
 
   if (rc == 0 &&
       write_err != HA_ERR_FOUND_DUPP_KEY &&
@@ -917,6 +938,7 @@ void select_union_recursive::cleanup()
     thd->rec_tables= tab;
     tbl->derived_result= 0;
   }
+  row_counter= 0;
 }
 
 
@@ -1687,7 +1709,8 @@ cont:
         Test if the aggregated data type is OK for a UNION element.
         E.g. in case of string data, DERIVATION_NONE is not allowed.
       */
-      if (type->type_handler()->union_element_finalize(type))
+      if (type->type() == Item::TYPE_HOLDER && type->type_handler()->
+            union_element_finalize(static_cast<Item_type_holder*>(type)))
         goto err;
     }
     
@@ -2561,7 +2584,8 @@ bool st_select_lex_unit::cleanup()
   {
     DBUG_RETURN(FALSE);
   }
-  if (with_element && with_element->is_recursive && union_result)
+  if (with_element && with_element->is_recursive && union_result &&
+      with_element->rec_outer_references)
   {
     select_union_recursive *result= with_element->rec_result;
     if (++result->cleanup_count == with_element->rec_outer_references)
@@ -2730,6 +2754,29 @@ static void cleanup_order(ORDER *order)
 }
 
 
+static void cleanup_window_funcs(List<Item_window_func> &win_funcs)
+{
+  List_iterator_fast<Item_window_func> it(win_funcs);
+  Item_window_func *win_func;
+  while ((win_func= it++))
+  {
+    Window_spec *win_spec= win_func->window_spec;
+    if (!win_spec)
+      continue;
+    if (win_spec->save_partition_list)
+    {
+      win_spec->partition_list= win_spec->save_partition_list;
+      win_spec->save_partition_list= NULL;
+    }
+    if (win_spec->save_order_list)
+    {
+      win_spec->order_list= win_spec->save_order_list;
+      win_spec->save_order_list= NULL;
+    }
+  }
+}
+
+
 bool st_select_lex::cleanup()
 {
   bool error= FALSE;
@@ -2741,6 +2788,8 @@ bool st_select_lex::cleanup()
   cleanup_order(group_list.first);
   cleanup_ftfuncs(this);
 
+  cleanup_window_funcs(window_funcs);
+
   if (join)
   {
     List_iterator<TABLE_LIST> ti(leaf_tables);
@@ -2767,7 +2816,8 @@ bool st_select_lex::cleanup()
   for (SELECT_LEX_UNIT *lex_unit= first_inner_unit(); lex_unit ;
        lex_unit= lex_unit->next_unit())
   {
-    if (lex_unit->with_element && lex_unit->with_element->is_recursive)
+    if (lex_unit->with_element && lex_unit->with_element->is_recursive &&
+        lex_unit->with_element->rec_outer_references)
       continue;
     error= (bool) ((uint) error | (uint) lex_unit->cleanup());
   }
