@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2014, 2019, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2020, MariaDB Corporation.
+Copyright (c) 2017, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -170,7 +170,8 @@ inline void PageBulk::insertPage(rec_t *rec, rec_offs *offsets)
       (fmt == REDUNDANT ? PAGE_OLD_INFIMUM : PAGE_NEW_INFIMUM))
   {
     const rec_t *old_rec = m_cur_rec;
-    rec_offs *old_offsets= rec_get_offsets(old_rec, m_index, nullptr, is_leaf,
+    rec_offs *old_offsets= rec_get_offsets(old_rec, m_index, nullptr, is_leaf
+                                           ? m_index->n_core_fields : 0,
                                            ULINT_UNDEFINED, &m_heap);
     ut_ad(cmp_rec_rec(rec, old_rec, offsets, old_offsets, m_index) > 0);
   }
@@ -361,7 +362,6 @@ inline void PageBulk::finishPage()
   ut_ad((fmt != REDUNDANT) == m_is_comp);
 
   ulint count= 0;
-  ulint n_recs= 0;
   byte *slot= my_assume_aligned<2>(m_page + srv_page_size -
                                    (PAGE_DIR + PAGE_DIR_SLOT_SIZE));
   const page_dir_slot_t *const slot0 = slot;
@@ -377,7 +377,6 @@ inline void PageBulk::finishPage()
       ut_ad(offset >= PAGE_NEW_SUPREMUM);
       ut_ad(offset < page_offset(slot));
       count++;
-      n_recs++;
 
       if (count == (PAGE_DIR_SLOT_MAX_N_OWNED + 1) / 2)
       {
@@ -431,7 +430,6 @@ inline void PageBulk::finishPage()
     while (insert_rec != m_page + PAGE_OLD_SUPREMUM)
     {
       count++;
-      n_recs++;
 
       if (count == (PAGE_DIR_SLOT_MAX_N_OWNED + 1) / 2)
       {
@@ -604,6 +602,7 @@ PageBulk::getSplitRec()
 
 	ut_ad(m_page_zip != NULL);
 	ut_ad(m_rec_no >= 2);
+	ut_ad(!m_index->is_instant());
 
 	ut_ad(page_get_free_space_of_empty(m_is_comp) > m_free_space);
 	total_used_size = page_get_free_space_of_empty(m_is_comp)
@@ -613,13 +612,13 @@ PageBulk::getSplitRec()
 	n_recs = 0;
 	offsets = NULL;
 	rec = page_get_infimum_rec(m_page);
+	const ulint n_core = page_is_leaf(m_page) ? m_index->n_core_fields : 0;
 
 	do {
 		rec = page_rec_get_next(rec);
 		ut_ad(page_rec_is_user_rec(rec));
 
-		offsets = rec_get_offsets(rec, m_index, offsets,
-					  page_is_leaf(m_page),
+		offsets = rec_get_offsets(rec, m_index, offsets, n_core,
 					  ULINT_UNDEFINED, &m_heap);
 		total_recs_size += rec_offs_size(offsets);
 		n_recs++;
@@ -648,9 +647,11 @@ PageBulk::copyIn(
 	ut_ad(m_rec_no == 0);
 	ut_ad(page_rec_is_user_rec(rec));
 
+	const ulint n_core = page_rec_is_leaf(rec)
+		? m_index->n_core_fields : 0;
+
 	do {
-		offsets = rec_get_offsets(rec, m_index, offsets,
-					  page_rec_is_leaf(split_rec),
+		offsets = rec_get_offsets(rec, m_index, offsets, n_core,
 					  ULINT_UNDEFINED, &m_heap);
 
 		insert(rec, offsets);
@@ -691,8 +692,10 @@ PageBulk::copyOut(
 	/* Set last record's next in page */
 	rec_offs*	offsets = NULL;
 	rec = page_rec_get_prev(split_rec);
-	offsets = rec_get_offsets(rec, m_index, offsets,
-				  page_rec_is_leaf(split_rec),
+	const ulint n_core = page_rec_is_leaf(split_rec)
+		? m_index->n_core_fields : 0;
+
+	offsets = rec_get_offsets(rec, m_index, offsets, n_core,
 				  ULINT_UNDEFINED, &m_heap);
 	mach_write_to_2(rec - REC_NEXT, m_is_comp
 			? static_cast<uint16_t>
@@ -703,8 +706,7 @@ PageBulk::copyOut(
 	m_cur_rec = rec;
 	m_heap_top = rec_get_end(rec, offsets);
 
-	offsets = rec_get_offsets(last_rec, m_index, offsets,
-				  page_rec_is_leaf(split_rec),
+	offsets = rec_get_offsets(last_rec, m_index, offsets, n_core,
 				  ULINT_UNDEFINED, &m_heap);
 
 	m_free_space += ulint(rec_get_end(last_rec, offsets) - m_heap_top)
@@ -815,14 +817,6 @@ PageBulk::storeExt(
 	dberr_t	err = btr_store_big_rec_extern_fields(
 		&btr_pcur, offsets, big_rec, &m_mtr, BTR_STORE_INSERT_BULK);
 
-	/* Reset m_block and m_cur_rec from page cursor, because
-	block may be changed during blob insert. (FIXME: Can it really?) */
-	ut_ad(m_block == btr_pcur.btr_cur.page_cur.block);
-
-	m_block = btr_pcur.btr_cur.page_cur.block;
-	m_cur_rec = btr_pcur.btr_cur.page_cur.rec;
-	m_page = buf_block_get_frame(m_block);
-
 	return(err);
 }
 
@@ -1118,7 +1112,8 @@ BtrBulk::insert(
 	/* Convert tuple to rec. */
         rec = rec_convert_dtuple_to_rec(static_cast<byte*>(mem_heap_alloc(
 		page_bulk->m_heap, rec_size)), m_index, tuple, n_ext);
-        offsets = rec_get_offsets(rec, m_index, offsets, !level,
+        offsets = rec_get_offsets(rec, m_index, offsets, level
+				  ? 0 : m_index->n_core_fields,
 				  ULINT_UNDEFINED, &page_bulk->m_heap);
 
 	page_bulk->insert(rec, offsets);
