@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2013, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2016, 2020, MariaDB Corporation.
+Copyright (c) 2016, 2022, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -47,7 +47,7 @@ SysTablespace srv_tmp_space;
 
 /** If the last data file is auto-extended, we add this many pages to it
 at a time. We have to make this public because it is a config variable. */
-ulong sys_tablespace_auto_extend_increment;
+uint sys_tablespace_auto_extend_increment;
 
 /** Convert a numeric string that optionally ends in G or M or K,
     to a number containing megabytes.
@@ -275,7 +275,8 @@ SysTablespace::parse_params(
 			}
 		}
 
-		m_files.push_back(Datafile(filepath, flags(), size, order));
+		m_files.push_back(Datafile(filepath, flags(), uint32_t(size),
+					   order));
 		Datafile* datafile = &m_files.back();
 		datafile->make_filepath(path(), filepath, NO_EXT);
 
@@ -351,7 +352,7 @@ SysTablespace::check_size(
 	also the data file could contain an incomplete extent.
 	So we need to round the size downward to a  megabyte.*/
 
-	const ulint	rounded_size_pages = static_cast<ulint>(
+	const uint32_t	rounded_size_pages = static_cast<uint32_t>(
 		size >> srv_page_size_shift);
 
 	/* If last file */
@@ -446,12 +447,27 @@ SysTablespace::create_file(
 
 	case SRV_NOT_RAW:
 		err = file.open_or_create(
-			m_ignore_read_only ? false : srv_read_only_mode);
+			!m_ignore_read_only && srv_read_only_mode);
 		break;
 	}
 
+	if (err != DB_SUCCESS) {
+		return err;
+	}
 
-	if (err == DB_SUCCESS && file.m_type != SRV_OLD_RAW) {
+	switch (file.m_type) {
+	case SRV_OLD_RAW:
+		break;
+	case SRV_NOT_RAW:
+#ifndef _WIN32
+		if (!space_id() && my_disable_locking
+		    && os_file_lock(file.m_handle, file.m_filepath)) {
+			err = DB_ERROR;
+			break;
+		}
+#endif
+		/* fall through */
+	case SRV_NEW_RAW:
 		err = set_size(file);
 	}
 
@@ -492,7 +508,7 @@ SysTablespace::open_file(
 
 	case SRV_NOT_RAW:
 		err = file.open_or_create(
-			m_ignore_read_only ? false : srv_read_only_mode);
+			!m_ignore_read_only && srv_read_only_mode);
 
 		if (err != DB_SUCCESS) {
 			return(err);
@@ -507,6 +523,14 @@ SysTablespace::open_file(
 		break;
 
 	case SRV_NOT_RAW:
+#ifndef _WIN32
+		if (!space_id() && (m_ignore_read_only || !srv_read_only_mode)
+		    && my_disable_locking
+		    && os_file_lock(file.m_handle, file.m_filepath)) {
+			err = DB_ERROR;
+			break;
+		}
+#endif
 		/* Check file size for existing file. */
 		err = check_size(file);
 		break;
@@ -559,7 +583,7 @@ SysTablespace::read_lsn_and_check_flags(lsn_t* flushed_lsn)
 	ut_a(it->order() == 0);
 
 	if (srv_operation == SRV_OPERATION_NORMAL) {
-		buf_dblwr_init_or_load_pages(it->handle(), it->filepath());
+		buf_dblwr.init_or_load_pages(it->handle(), it->filepath());
 	}
 
 	/* Check the contents of the first page of the
@@ -678,13 +702,18 @@ SysTablespace::file_not_found(
 {
 	file.m_exists = false;
 
-	if (srv_read_only_mode && !m_ignore_read_only) {
+	if (m_ignore_read_only) {
+	} else if (srv_read_only_mode) {
 		ib::error() << "Can't create file '" << file.filepath()
 			<< "' when --innodb-read-only is set";
-
 		return(DB_ERROR);
+	} else if (srv_force_recovery && space_id() == TRX_SYS_SPACE) {
+		ib::error() << "Can't create file '" << file.filepath()
+			<< "' when --innodb-force-recovery is set";
+		return DB_ERROR;
+	}
 
-	} else if (&file == &m_files.front()) {
+	if (&file == &m_files.front()) {
 
 		/* First data file. */
 		ut_a(!*create_new_db);
@@ -904,30 +933,22 @@ SysTablespace::open_or_create(
 
 		if (it != begin) {
 		} else if (is_temp) {
-			ut_ad(!fil_system.temp_space);
 			ut_ad(space_id() == SRV_TMP_SPACE_ID);
-			space = fil_space_create(
+			space = fil_space_t::create(
 				name(), SRV_TMP_SPACE_ID, flags(),
 				FIL_TYPE_TEMPORARY, NULL);
-
-			mutex_enter(&fil_system.mutex);
-			fil_system.temp_space = space;
-			mutex_exit(&fil_system.mutex);
+			ut_ad(space == fil_system.temp_space);
 			if (!space) {
 				return DB_ERROR;
 			}
 			ut_ad(!space->is_compressed());
 			ut_ad(space->full_crc32());
 		} else {
-			ut_ad(!fil_system.sys_space);
 			ut_ad(space_id() == TRX_SYS_SPACE);
-			space = fil_space_create(
+			space = fil_space_t::create(
 				name(), TRX_SYS_SPACE, it->flags(),
 				FIL_TYPE_TABLESPACE, NULL);
-
-			mutex_enter(&fil_system.mutex);
-			fil_system.sys_space = space;
-			mutex_exit(&fil_system.mutex);
+			ut_ad(space == fil_system.sys_space);
 			if (!space) {
 				return DB_ERROR;
 			}
@@ -935,10 +956,10 @@ SysTablespace::open_or_create(
 
 		ut_a(fil_validate());
 
-		ulint	max_size = (++node_counter == m_files.size()
+		uint32_t max_size = (++node_counter == m_files.size()
 				    ? (m_last_file_size_max == 0
-				       ? ULINT_MAX
-				       : m_last_file_size_max)
+				       ? UINT32_MAX
+				       : uint32_t(m_last_file_size_max))
 				    : it->m_size);
 
 		space->add(it->m_filepath, OS_FILE_CLOSED, it->m_size,
@@ -965,30 +986,21 @@ SysTablespace::normalize_size()
 
 /**
 @return next increment size */
-ulint
-SysTablespace::get_increment() const
+uint32_t SysTablespace::get_increment() const
 {
-	ulint	increment;
+  if (m_last_file_size_max == 0)
+    return get_autoextend_increment();
 
-	if (m_last_file_size_max == 0) {
-		increment = get_autoextend_increment();
-	} else {
+  if (!is_valid_size())
+  {
+     ib::error() << "The last data file in " << name()
+                 << " has a size of " << last_file_size()
+                 << " but the max size allowed is "
+                 << m_last_file_size_max;
+  }
 
-		if (!is_valid_size()) {
-			ib::error() << "The last data file in " << name()
-				<< " has a size of " << last_file_size()
-				<< " but the max size allowed is "
-				<< m_last_file_size_max;
-		}
-
-		increment = m_last_file_size_max - last_file_size();
-	}
-
-	if (increment > get_autoextend_increment()) {
-		increment = get_autoextend_increment();
-	}
-
-	return(increment);
+  return std::min(uint32_t(m_last_file_size_max) - last_file_size(),
+                  get_autoextend_increment());
 }
 
 
