@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2012, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2019, MariaDB Corporation.
+Copyright (c) 2017, 2022, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -28,19 +28,12 @@ Created Apr 26, 2012 Vasil Dimov
 #define dict0stats_bg_h
 
 #include "dict0types.h"
-#include "os0event.h"
 #include "os0thread.h"
 
-
 #ifdef HAVE_PSI_INTERFACE
 extern mysql_pfs_key_t	dict_stats_recalc_pool_mutex_key;
 #endif /* HAVE_PSI_INTERFACE */
 
-#ifdef UNIV_DEBUG
-/** Value of MySQL global used to disable dict_stats thread. */
-extern my_bool		innodb_dict_stats_disabled_debug;
-#endif /* UNIV_DEBUG */
-
 /*****************************************************************//**
 Delete a given table from the auto recalc pool.
 dict_stats_recalc_pool_del() */
@@ -104,14 +97,6 @@ Free resources allocated by dict_stats_thread_init(), must be called
 after dict_stats task has exited. */
 void dict_stats_deinit();
 
-#ifdef UNIV_DEBUG
-/** Disables dict stats thread. It's used by:
-	SET GLOBAL innodb_dict_stats_disabled_debug = 1 (0).
-@param[in]	save		immediate result from check function */
-void dict_stats_disabled_debug_update(THD*, st_mysql_sys_var*, void*,
-				      const void* save);
-#endif /* UNIV_DEBUG */
-
 /** Start the dict stats timer. */
 void dict_stats_start();
 
